import onnx
import torch
from onnx.numpy_helper import to_array

from onnx_pytorch.op_code_generators import OpCodeGenerator


class UnsqueezeOpCodeGenerator(OpCodeGenerator):

  def __init__(self,
               onnx_ver=onnx.defs.onnx_opset_version(),
               torch_ver=torch.__version__):
    super(UnsqueezeOpCodeGenerator, self).__init__(onnx_ver, torch_ver)

  def gen(self, node, value_infos, initializers, rename_helper, tensor_inplace):
    attr_value_dict = self.get_attr_value_dict(node)
    inputs_str, outputs_str = self.gen_input_output_string(
        node, initializers, rename_helper, tensor_inplace)
    axes = attr_value_dict.get("axes", [])
    if len(node.input) == 2:
      assert node.input[
          1] in initializers, "Currently UnsqueezeOpCodeGenerator only support all of [axes] is in initializers."
      axes = to_array(initializers[node.input[1]])
    init_str, forward_str = [], []
    curr_input = inputs_str[0]
    for a in axes:
      forward_str.append(
          f"{outputs_str[0]} = torch.unsqueeze({curr_input}, {a})")
      curr_input = outputs_str[0]

    return {"init": init_str, "forward": forward_str}



from tempfile import TemporaryDirectory

import os
import importlib.util

import numpy as np
import onnx
import onnxruntime
from onnx.numpy_helper import to_array
from onnxruntime.tools.symbolic_shape_infer import SymbolicShapeInference
import pytest
import torch

from onnx_model_maker import *
from onnx_model_maker.ops import *
from onnx_pytorch import code_gen

torch.set_printoptions(8)


class TestModel:

  def _run(self, inputs_np, onnx_model):
    model = onnx.ModelProto()
    model.CopyFrom(onnx_model)
    sess_options = onnxruntime.SessionOptions()
    session = onnxruntime.InferenceSession(model.SerializeToString(),
                                           sess_options)
    ort_outputs = session.run(None, {k: v for k, v in inputs_np})
    model.graph.ClearField("value_info")
    model = SymbolicShapeInference.infer_shapes(model, 2**31 - 1, True, True, 1)
    with TemporaryDirectory() as tmpdir:
      tmpdir = "/Users/wenhao/Projects/onnx-pytorch/onnx_pytorch/tests/tmp"
      code_gen.gen(model,
                   output_dir=tmpdir,
                   tensor_inplace=True,
                   simplify_names=True)
      spec = importlib.util.spec_from_file_location(
          "model", os.path.join(tmpdir, "model.py"))
      mod = importlib.util.module_from_spec(spec)
      spec.loader.exec_module(mod)
      pt_outputs = mod.test_run_model(
          [torch.from_numpy(v) for _, v in inputs_np])
      assert np.allclose(ort_outputs, [o.detach().numpy() for o in pt_outputs],
                         atol=1e-5,
                         rtol=1e-5,
                         equal_nan=True)

  # def test_ugmarketing(self):
  #   ugmarkering_model_path = "/Users/wenhao/Projects/ai_speed/saved_model/model.onnx"
  #   if not os.path.exists(ugmarkering_model_path):
  #     return False
  #   model = onnx.load(ugmarkering_model_path)
  #   self._run(
  #       [("import/whole_input:0", np.random.randn(1, 2905).astype(np.float32))],
  #       model)
  #
  # def test_gongxian(self):
  #   model_path = "/Users/wenhao/Projects/ai_speed/cenaniang_new.onnx"
  #   if not os.path.exists(model_path):
  #     return False
  #   model = onnx.load(model_path)
  #   self._run([("input", np.random.randn(1, 4576).astype(np.float32))], model)

  # def test_dcn(self):
  #   model_path = "/Users/wenhao/Projects/DeepCTR/model.onnx"
  #   if not os.path.exists(model_path):
  #     return False
  #   model = onnx.load(model_path)
  #   b = 10
  #   self._run([
  #       ("x_5:0", np.random.randint(0, 188, (b, 1)).astype(np.int32)),
  #       ("x_4:0", np.random.randint(0, 20, (b, 1)).astype(np.int32)),
  #       ("x_3:0", np.random.randint(0, 7, (b, 1)).astype(np.int32)),
  #       ("x_2:0", np.random.randint(0, 2, (b, 1)).astype(np.int32)),
  #       ("x_1:0", np.random.randint(0, 193, (b, 1)).astype(np.int32)),
  #       ("x:0", np.random.randint(0, 187, (b, 1)).astype(np.int32)),
  #   ], model)

  def test_dcn_criteo(self):
    model_path = "/Users/wenhao/Projects/DeepCTR/dcn_criteo.onnx"
    if not os.path.exists(model_path):
      return False
    model = onnx.load(model_path)
    initializers = {i.name: i for i in model.graph.initializer}
    inputs = [i.name for i in model.graph.input if i.name not in initializers]
    nodes = {
        n.input[1]: n
        for n in model.graph.node
        if n.op_type == "Gather" and n.input[1] in inputs
    }
    mapping = {i: nodes[i].input[0].split("/")[1].split("_")[-1] for i in inputs if i in nodes}
    mapping.update({i: f"I{int(i.split('_')[1].split(':')[0]) - 25}" for i in inputs if i not in nodes})
    b = 10
    inputs_np = [
        (i,
         np.random.randint(0,
                           to_array(initializers[nodes[i].input[0]]).shape[0],
                           (b, 1)).astype(np.int32)
         if i in nodes else np.random.randn(b, 1).astype(np.float32))
        for i in inputs
    ]
    self._run(inputs_np, model)


if __name__ == '__main__':
  pytest.main(['-s', 'test_onnx_model_zoo.py'])



import logging
from tempfile import TemporaryDirectory

import os
import importlib.util

import numpy as np
import onnx
import onnxruntime
from onnx.helper import make_tensor
from onnxruntime.tools.symbolic_shape_infer import SymbolicShapeInference
import pytest
import torch

from onnx_model_maker import *
from onnx_model_maker.ops import *
from onnx_pytorch import code_gen

torch.set_printoptions(8)


class TestBase:

  def _run(self, inputs_np):
    model = onnx.ModelProto()
    model.CopyFrom(omm.model)
    sess_options = onnxruntime.SessionOptions()
    session = onnxruntime.InferenceSession(model.SerializeToString(),
                                           sess_options)
    ort_outputs = session.run(None, {k: v for k, v in inputs_np if k != ""})
    model.graph.ClearField("value_info")
    try:
      model = SymbolicShapeInference.infer_shapes(model, 2**31 - 1, True, True,
                                                  1)
    except:
      logging.warning("Shape infer by onnxruntime failed.")
    with TemporaryDirectory() as tmpdir:
      code_gen.gen(model,
                   output_dir=tmpdir,
                   tensor_inplace=True,
                   simplify_names=True)
      spec = importlib.util.spec_from_file_location(
          "model", os.path.join(tmpdir, "model.py"))
      mod = importlib.util.module_from_spec(spec)
      spec.loader.exec_module(mod)
      pt_outputs = mod.test_run_model(
          [torch.from_numpy(v) for k, v in inputs_np if k != ""])
      if type(pt_outputs) == torch.Tensor:
        pt_outputs = [pt_outputs.detach().numpy()]
      elif type(pt_outputs) in (list, tuple):
        pt_outputs = [o.detach().numpy() for o in pt_outputs]
      for l, r in zip(ort_outputs, pt_outputs):
        assert np.allclose(l, r, atol=1e-4, rtol=1e-4, equal_nan=True)

  def test_conv_flatten_relu(self):
    reset_model(13)
    nps = [np.random.randn(1, 3, 224, 224).astype(np.float32)]
    inputs = Input(*nps)
    conv_node = Conv(inputs[0],
                     np.random.randn(32, 3, 3, 3).astype(np.float32),
                     np.random.randn(32).astype(np.float32))
    flatten_node = Flatten(conv_node)
    relu_node = Relu(flatten_node)
    Output(relu_node)
    self._run(list(zip(inputs, nps)))

  def test_conv_batchnorm_maxpool_flatten_add_relu(self):
    reset_model(13)
    nps = [np.random.randn(1, 3, 224, 224).astype(np.float32)]
    inputs = Input(*nps)
    conv_node = Conv(inputs[0],
                     np.random.randn(32, 3, 3, 3).astype(np.float32),
                     np.random.randn(32).astype(np.float32))
    bn_node = BatchNormalization(
        conv_node,
        np.ones(32,).astype(np.float32),
        np.zeros(32,).astype(np.float32),
        np.random.randn(32).astype(np.float32),
        np.random.randn(32).astype(np.float32),
    )
    max_pool_node = MaxPool(bn_node,
                            kernel_shape=(3, 3),
                            strides=(2, 2),
                            pads=(0, 0, 1, 1))
    flatten_node = Flatten(max_pool_node, axis=1)
    add_node = Add(flatten_node, np.random.randn(1).astype(np.float32))
    relu_node = Relu(add_node)
    Output(relu_node)
    self._run(list(zip(inputs, nps)))

  def test_abs(self):
    reset_model(13)
    nps = [np.random.randn(1, 10).astype(np.float32)]
    inputs = Input(*nps)
    Output(Abs(inputs[0]))
    self._run(list(zip(inputs, nps)))

  def test_acos(self):
    reset_model(13)
    nps = [np.random.randn(5).astype(np.float32)]
    inputs = Input(*nps)
    Output(Acos(inputs[0]))
    self._run(list(zip(inputs, nps)))

  def test_acosh(self):
    reset_model(13)
    nps = [np.random.randn(5).astype(np.float32)]
    inputs = Input(*nps)
    Output(Acosh(inputs[0]))
    self._run(list(zip(inputs, nps)))

  def test_add(self):
    reset_model(13)
    nps = [np.random.randn(1, 10).astype(np.float32)]
    inputs = Input(*nps)
    Output(Add(inputs[0], np.random.randn(1, 10).astype(np.float32)))
    self._run(list(zip(inputs, nps)))

  def test_and(self):
    reset_model(13)
    nps = [
        np.random.randint(low=0, high=1, size=(5,)).astype(np.bool),
        np.random.randint(low=0, high=1, size=(5,)).astype(np.bool)
    ]
    inputs = Input(*nps)
    Output(And(*inputs))
    self._run(list(zip(inputs, nps)))

  def test_argmax(self):
    reset_model(13)
    nps = [np.random.randn(1, 10).astype(np.float32)]
    inputs = Input(*nps)
    Output(ArgMax(inputs, axis=1))
    self._run(list(zip(inputs, nps)))

  def test_argmin(self):
    reset_model(13)
    nps = [np.random.randn(1, 10).astype(np.float32)]
    inputs = Input(*nps)
    Output(ArgMin(inputs, axis=1))
    self._run(list(zip(inputs, nps)))

  def test_asin(self):
    reset_model(13)
    nps = [np.random.randn(5).astype(np.float32)]
    inputs = Input(*nps)
    Output(Asin(inputs[0]))
    self._run(list(zip(inputs, nps)))

  def test_asinh(self):
    reset_model(13)
    nps = [np.random.randn(5).astype(np.float32)]
    inputs = Input(*nps)
    Output(Asinh(inputs[0]))
    self._run(list(zip(inputs, nps)))

  def test_atan(self):
    reset_model(13)
    nps = [np.random.randn(5).astype(np.float32)]
    inputs = Input(*nps)
    Output(Atan(inputs[0]))
    self._run(list(zip(inputs, nps)))

  def test_atanh(self):
    reset_model(13)
    nps = [np.random.randn(5).astype(np.float32)]
    inputs = Input(*nps)
    Output(Atanh(inputs[0]))
    self._run(list(zip(inputs, nps)))

  def test_avg_pool(self):
    reset_model(13)
    nps = [np.random.randn(1, 1, 5, 5).astype(np.float32)]
    inputs = Input(*nps)
    Output(AveragePool(inputs, kernel_shape=(3, 3), pads=(0, 0, 1, 1)))
    self._run(list(zip(inputs, nps)))

  def test_avg_pool_no_pad(self):
    reset_model(13)
    nps = [np.random.randn(1, 1, 6, 6).astype(np.float32)]
    inputs = Input(*nps)
    Output(AveragePool(inputs, kernel_shape=(3, 3)))
    self._run(list(zip(inputs, nps)))

  def test_bitshift_left(self):
    reset_model(13)
    nps = [np.array([1, 2]).astype(np.uint8), np.array([1, 2]).astype(np.uint8)]
    inputs = Input(*nps)
    Output(BitShift(*inputs, direction="LEFT"))
    self._run(list(zip(inputs, nps)))

  def test_bitshift_right(self):
    reset_model(13)
    nps = [np.array([1, 4]).astype(np.uint8), np.array([1, 1]).astype(np.uint8)]
    inputs = Input(*nps)
    Output(BitShift(*inputs, direction="RIGHT"))
    self._run(list(zip(inputs, nps)))

  def test_batch_normalization(self):
    reset_model(13)
    nps = [np.random.randn(1, 32, 3, 3).astype(np.float32)]
    inputs = Input(*nps)
    Output(
        BatchNormalization(
            inputs[0],
            np.ones(32,).astype(np.float32),
            np.zeros(32,).astype(np.float32),
            np.random.randn(32).astype(np.float32),
            np.random.randn(32).astype(np.float32),
        ))
    self._run(list(zip(inputs, nps)))

  def test_cast(self):
    reset_model(13)
    nps = [
        np.random.randn(1, 10).astype(np.float32),
    ]
    inputs = Input(*nps)
    Output(Cast(inputs, to=6))
    self._run(list(zip(inputs, nps)))

  def test_ceil(self):
    reset_model(13)
    nps = [np.random.randn(1, 10).astype(np.float32)]
    inputs = Input(*nps)
    Output(Ceil(inputs[0]))
    self._run(list(zip(inputs, nps)))

  def test_clip(self):
    reset_model(13)
    nps = [
        np.random.randn(1, 5).astype(np.float32),
        np.asarray(0).astype(np.float32)
    ]
    inputs = Input(*nps)
    Output(Clip(*inputs))
    self._run(list(zip(inputs, nps)))

  def test_concat(self):
    reset_model(13)
    nps = [
        np.random.randn(1, 10).astype(np.float32),
        np.random.randn(2, 10).astype(np.float32)
    ]
    inputs = Input(*nps)
    Output(Concat(inputs, axis=0))
    self._run(list(zip(inputs, nps)))

  def test_constant_add(self):
    reset_model(13)
    nps = [
        np.random.randn(1, 10).astype(np.float32),
    ]
    inputs = Input(*nps)
    constant_value = np.random.randn(1, 10).astype(np.float32)
    t = make_tensor("", 1, constant_value.shape, constant_value.flatten())
    Output(Add(inputs, Constant(value=t)))
    self._run(list(zip(inputs, nps)))

  def test_constant_of_shape(self):
    reset_model(13)
    nps = [
        np.array([2, 3]).astype(np.int64),
    ]
    inputs = Input(*nps)
    constant_value = np.random.randn(1).astype(np.float32)
    t = make_tensor("", 1, constant_value.shape, constant_value)
    Output(ConstantOfShape(inputs, value=t))
    self._run(list(zip(inputs, nps)))

  def test_conv(self):
    reset_model(13)
    nps = [np.random.randn(1, 3, 3, 3).astype(np.float32)]
    inputs = Input(*nps)
    Output(
        Conv(
            inputs[0],
            np.random.randn(32, 3, 3, 3).astype(np.float32),
            np.random.randn(32).astype(np.float32),
        ))
    self._run(list(zip(inputs, nps)))

  def test_conv_transpose(self):
    reset_model(13)
    nps = [np.random.randn(1, 3, 3, 3).astype(np.float32)]
    inputs = Input(*nps)
    Output(
        ConvTranspose(
            inputs[0],
            np.random.randn(3, 32, 3, 3).astype(np.float32),
            np.random.randn(32).astype(np.float32),
        ))
    self._run(list(zip(inputs, nps)))

  def test_conv_transpose_pads(self):
    reset_model(13)
    nps = [np.random.randn(1, 1, 3, 3).astype(np.float32)]
    inputs = Input(*nps)
    Output(
        ConvTranspose(inputs[0],
                      np.random.randn(1, 2, 3, 3).astype(np.float32),
                      strides=[3, 2],
                      pads=[1, 2, 1, 2]))
    self._run(list(zip(inputs, nps)))

  def test_conv_transpose_dilations(self):
    reset_model(13)
    nps = [np.random.randn(1, 1, 3, 3).astype(np.float32)]
    inputs = Input(*nps)
    Output(
        ConvTranspose(
            inputs[0],
            np.random.randn(1, 1, 2, 2).astype(np.float32),
            dilations=[2, 2],
        ))
    self._run(list(zip(inputs, nps)))

  def test_conv_transpose_attributes(self):
    reset_model(13)
    nps = [np.random.randn(1, 1, 3, 3).astype(np.float32)]
    inputs = Input(*nps)
    Output(
        ConvTranspose(
            inputs[0],
            np.random.randn(1, 2, 3, 3).astype(np.float32),
            strides=[3, 2],
            output_shape=[10, 8],
            output_padding=[1, 1],
        ))
    self._run(list(zip(inputs, nps)))

  def test_cos(self):
    reset_model(13)
    nps = [np.random.randn(5).astype(np.float32)]
    inputs = Input(*nps)
    Output(Cos(inputs[0]))
    self._run(list(zip(inputs, nps)))

  def test_cosh(self):
    reset_model(13)
    nps = [np.random.randn(5).astype(np.float32)]
    inputs = Input(*nps)
    Output(Cosh(inputs[0]))
    self._run(list(zip(inputs, nps)))

  def test_flatten(self):
    reset_model(13)
    nps = [np.random.randn(1, 3, 3, 3).astype(np.float32)]
    inputs = Input(*nps)
    Output(Flatten(inputs))
    self._run(list(zip(inputs, nps)))

  def test_gather(self):
    reset_model(13)
    nps = [
        np.random.randn(6, 8, 3).astype(np.float32),
        np.random.randn(*[3, 2]).astype(np.int64),
    ]
    inputs = Input(*nps)
    Output(Gather(*inputs, axis=0))
    self._run(list(zip(inputs, nps)))

  def test_gather_axis_1(self):
    reset_model(13)
    nps = [
        np.random.randn(6, 3, 3).astype(np.float32),
        np.array([[0, 2], [0, 1], [2, 0]]).astype(np.int64),
    ]
    inputs = Input(*nps)
    Output(Gather(*inputs, axis=1))
    self._run(list(zip(inputs, nps)))

  def test_gather_nd(self):
    reset_model(13)
    nps = [
        np.random.randn(6, 8, 3).astype(np.float32),
        np.random.randn(*[3, 2]).astype(np.int64),
    ]
    inputs = Input(*nps)
    Output(GatherND(*inputs, batch_dims=0))
    self._run(list(zip(inputs, nps)))

  @pytest.mark.skip(reason="Not implemented for batch_dims != 0")
  def test_gather_nd_batch_dims_1(self):
    reset_model(13)
    nps = [
        np.array([[[0, 1], [2, 3]], [[4, 5], [6, 7]]]),
        np.array([[1], [0]]).astype(np.int64),
    ]
    inputs = Input(*nps)
    Output(GatherND(*inputs, batch_dims=1))
    self._run(list(zip(inputs, nps)))

  def test_gemm(self):
    reset_model(13)
    nps = [
        np.random.randn(2, 3).astype(np.float32),
        np.random.randn(3, 4).astype(np.float32),
        np.random.randn(2, 4).astype(np.float32),
    ]
    inputs = Input(*nps)
    Output(Gemm(*inputs))
    self._run(list(zip(inputs, nps)))

  def test_global_average_pool(self):
    reset_model(13)
    nps = [np.random.randn(2, 3, 4).astype(np.float32)]
    inputs = Input(*nps)
    Output(GlobalAveragePool(*inputs))
    self._run(list(zip(inputs, nps)))

  def test_mat_mul(self):
    reset_model(13)
    nps = [
        np.random.randn(5, 2, 3).astype(np.float32),
        np.random.randn(5, 3, 2).astype(np.float32)
    ]
    inputs = Input(*nps)
    Output(MatMul(*inputs))
    self._run(list(zip(inputs, nps)))

  def test_max(self):
    reset_model(13)
    nps = [
        np.random.randn(1, 10).astype(np.float32),
        np.random.randn(2, 10).astype(np.float32)
    ]
    inputs = Input(*nps)
    Output(Max(inputs))
    self._run(list(zip(inputs, nps)))

  def test_max_pool(self):
    reset_model(13)
    nps = [np.random.randn(1, 1, 5, 5).astype(np.float32)]
    inputs = Input(*nps)
    Output(MaxPool(inputs, kernel_shape=(3, 3), pads=(0, 0, 1, 1)))
    self._run(list(zip(inputs, nps)))

  def test_mul(self):
    reset_model(13)
    nps = [
        np.random.randn(1, 2, 3).astype(np.float32),
        np.random.randn(1, 2, 3).astype(np.float32)
    ]
    inputs = Input(*nps)
    Output(Mul(*inputs))
    self._run(list(zip(inputs, nps)))

  def test_pad(self):
    reset_model(13)
    nps = [
        np.random.randn(1, 2, 3).astype(np.float32),
    ]
    inputs = Input(*nps)
    Output(Pad(*inputs, np.array([0, 0, 1, 0, 0, 2])))
    self._run(list(zip(inputs, nps)))

  def test_pad_5D(self):
    reset_model(13)
    nps = [
        np.random.randn(1, 2, 3, 4, 5).astype(np.float32),
    ]
    inputs = Input(*nps)
    Output(
        Pad(*inputs, np.array([0, 0, 1, 2, 3, 0, 0, 2, 3, 4]),
            np.array([1.0]).astype(np.float32)))
    self._run(list(zip(inputs, nps)))

  def test_pad_reflect(self):
    reset_model(13)
    nps = [
        np.random.randn(1, 2, 3, 4).astype(np.float32),
    ]
    inputs = Input(*nps)
    Output(
        Pad(*inputs,
            np.array([0, 0, 1, 2, 0, 0, 2, 3]),
            np.array([1.0]).astype(np.float32),
            mode="reflect"))
    self._run(list(zip(inputs, nps)))

  def test_reciprocal(self):
    reset_model(13)
    nps = [np.random.randn(1, 10).astype(np.float32)]
    inputs = Input(*nps)
    Output(Reciprocal(inputs[0]))
    self._run(list(zip(inputs, nps)))

  def test_reduce_prod(self):
    reset_model(13)
    nps = [np.random.randn(1, 2, 3).astype(np.float32)]
    inputs = Input(*nps)
    Output(ReduceProd(inputs[0], axes=np.array((1, 2)).astype(np.int64)))
    self._run(list(zip(inputs, nps)))

  def test_reduce_sum(self):
    reset_model(13)
    nps = [np.random.randn(1, 2, 3).astype(np.float32)]
    inputs = Input(*nps)
    Output(ReduceSum(inputs[0], np.array((1, 2)).astype(np.int64)))
    self._run(list(zip(inputs, nps)))

  def test_relu(self):
    reset_model(13)
    nps = [np.random.randn(1, 5).astype(np.float32)]
    inputs = Input(*nps)
    Output(Relu(inputs))
    self._run(list(zip(inputs, nps)))

  def test_reshape(self):
    reset_model(13)
    nps = [np.random.randn(4,).astype(np.float32)]
    inputs = Input(*nps)
    Output(Reshape(inputs[0], np.array((2, 2)).astype(np.int64)))
    self._run(list(zip(inputs, nps)))

  def test_resize_scales_nearest(self):
    reset_model(13)
    nps = [
        np.array([[[
            [1, 2, 3, 4],
            [5, 6, 7, 8],
        ]]], dtype=np.float32),
        np.array([], dtype=np.float32),
    ]
    inputs = Input(*nps)
    Output(
        Resize(
            *inputs,
            np.array([1.0, 1.0, 0.6, 0.6], dtype=np.float32),
            mode="nearest",
        ))
    self._run(list(zip(inputs, nps)))

  def test_resize_downsample_sizes_linear_pytorch_half_pixel(self):
    reset_model(13)
    nps = [
        np.array([[[
            [1, 2, 3, 4],
            [5, 6, 7, 8],
            [9, 10, 11, 12],
            [13, 14, 15, 16],
        ]]],
                 dtype=np.float32),
        np.array([], dtype=np.float32),
        np.array([], dtype=np.float32),
    ]
    inputs = Input(*nps)
    Output(
        Resize(*inputs,
               np.array([1, 1, 3, 1], dtype=np.int64),
               mode='linear',
               coordinate_transformation_mode='pytorch_half_pixel'))
    self._run(list(zip(inputs, nps)))

  def test_resize_pt_nearest(self):
    reset_model(13)
    nps = [
        np.array([[[[1., 2., 0.], [3., 4., 0.], [0., 0., 0.]]]],
                 dtype=np.float32),
        np.array([], dtype=np.float32),
    ]
    inputs = Input(*nps)
    Output(
        Resize(
            *inputs,
            np.array([1.0, 1.0, 2.0, 2.0], dtype=np.float32),
            mode="nearest",
        ))
    self._run(list(zip(inputs, nps)))

  def test_resize_pt_bilinear(self):
    reset_model(13)
    nps = [
        np.array([[[[1., 2., 0.], [3., 4., 0.], [0., 0., 0.]]]],
                 dtype=np.float32),
        np.array([], dtype=np.float32),
    ]
    inputs = Input(*nps)
    Output(
        Resize(
            *inputs,
            np.array([1.0, 1.0, 2.0, 2.0], dtype=np.float32),
            mode="linear",
        ))
    self._run(list(zip(inputs, nps)))

  def test_resize_pt_bilinear_align_corners(self):
    reset_model(13)
    nps = [
        np.array([[[[1., 2., 0.], [3., 4., 0.], [0., 0., 0.]]]],
                 dtype=np.float32),
        np.array([], dtype=np.float32),
    ]
    inputs = Input(*nps)
    Output(
        Resize(
            *inputs,
            np.array([1.0, 1.0, 2.0, 2.0], dtype=np.float32),
            mode="linear",
            coordinate_transformation_mode="align_corners",
        ))
    self._run(list(zip(inputs, nps)))

  def test_shape(self):
    reset_model(13)
    nps = [np.random.randn(1, 2, 3, 4).astype(np.float32)]
    inputs = Input(*nps)
    Output(Shape(inputs[0]))
    self._run(list(zip(inputs, nps)))

  def test_sigmoid(self):
    reset_model(13)
    nps = [np.random.randn(1, 10).astype(np.float32)]
    inputs = Input(*nps)
    Output(Sigmoid(inputs[0]))
    self._run(list(zip(inputs, nps)))

  def test_slice(self):
    reset_model(13)
    nps = [np.random.randn(5, 5).astype(np.float32)]
    inputs = Input(*nps)
    Output(
        Slice(inputs[0],
              np.array((2, 3)).astype(np.int64),
              np.array((4, 5)).astype(np.int64)))
    self._run(list(zip(inputs, nps)))

  def test_softmax(self):
    reset_model(13)
    nps = [np.random.randn(1, 10).astype(np.float32)]
    inputs = Input(*nps)
    Output(Softmax(inputs[0]))
    self._run(list(zip(inputs, nps)))

  def test_split(self):
    reset_model(13)
    nps = [np.random.randn(1, 10).astype(np.float32)]
    inputs = Input(*nps)
    Output(Split(inputs, split=np.array([2, 8]), axis=1))
    self._run(list(zip(inputs, nps)))

  def test_sqrt(self):
    reset_model(13)
    nps = [np.random.randn(1, 10).astype(np.float32)]
    inputs = Input(*nps)
    Output(Sqrt(inputs[0]))
    self._run(list(zip(inputs, nps)))

  def test_squeeze(self):
    reset_model(13)
    nps = [np.random.randn(1, 10, 1, 1).astype(np.float32)]
    inputs = Input(*nps)
    Output(Squeeze(inputs[0], np.array(([2, 3]))))
    self._run(list(zip(inputs, nps)))

  def test_squeeze_no_axes(self):
    reset_model(13)
    nps = [np.random.randn(1, 10, 1, 1).astype(np.float32)]
    inputs = Input(*nps)
    Output(Squeeze(inputs[0]))
    self._run(list(zip(inputs, nps)))

  def test_transpose(self):
    reset_model(13)
    nps = [np.random.randn(1, 2, 3, 4).astype(np.float32)]
    inputs = Input(*nps)
    Output(Transpose(inputs[0], perm=[0, 2, 3, 1]))
    self._run(list(zip(inputs, nps)))

  def test_transpose_no_perm(self):
    reset_model(13)
    nps = [np.random.randn(1, 2, 3, 4).astype(np.float32)]
    inputs = Input(*nps)
    Output(Transpose(inputs[0]))
    self._run(list(zip(inputs, nps)))

  def test_unsqueeze(self):
    reset_model(13)
    nps = [np.random.randn(1, 2).astype(np.float32)]
    inputs = Input(*nps)
    Output(Unsqueeze(inputs[0], np.array(([2, 3]))))
    self._run(list(zip(inputs, nps)))


if __name__ == '__main__':
  pytest.main(['-s', 'test_base.py'])



import onnx
import torch

from onnx_pytorch.op_code_generators import OpCodeGenerator


class AveragePoolOpCodeGenerator(OpCodeGenerator):

  def __init__(self,
               onnx_ver=onnx.defs.onnx_opset_version(),
               torch_ver=torch.__version__):
    super(AveragePoolOpCodeGenerator, self).__init__(onnx_ver, torch_ver)

  def gen(self, node, value_infos, initializers, rename_helper, tensor_inplace):
    attr_value_dict = self.get_attr_value_dict(node)
    inputs_str, outputs_str = self.gen_input_output_string(
        node, initializers, rename_helper, tensor_inplace)

    d = len(value_infos[node.input[0]].type.tensor_type.shape.dim) - 2
    assert (d in (1, 2, 3))

    nn_name = f"AvgPool{d}d"
    node_name = rename_helper.get_node_name(node.name, node.op_type)
    init_str, forward_str = [], []
    slice_str = [":", ":"]

    param = {
        "kernel_size": attr_value_dict["kernel_shape"][:].__repr__(),
        "ceil_mode": bool(attr_value_dict["ceil_mode"]),
        "stride": attr_value_dict.get("strides", 1),
        "count_include_pad": bool(attr_value_dict.get("count_include_pad", 0))
    }
    if "pads" in attr_value_dict:
      padding = []
      for i in range(d):
        padding_size = max(attr_value_dict['pads'][i],
                           attr_value_dict['pads'][i + d])
        padding.append(padding_size)
        slice_begin = "" if padding_size == attr_value_dict['pads'][i] else str(
            padding_size - attr_value_dict['pads'][i])
        slice_end = "" if padding_size == attr_value_dict['pads'][
            i + d] else str(attr_value_dict['pads'][i + d] - padding_size)
        slice_str.append(":".join([slice_begin, slice_end]))
      param["padding"] = padding.__repr__()
    params_str = self.gen_params_str(**param,)
    init_str.append(f"self.{node_name} = nn.{nn_name}(**{{{params_str}}})")
    forward_str.append(
        f"{outputs_str[0]} = self.{node_name}({inputs_str[0]})[{', '.join(slice_str)}]"
    )

    return {"init": init_str, "forward": forward_str}



import onnx
import torch

from onnx_pytorch.op_code_generators import OpCodeGenerator


class ResizeOpCodeGenerator(OpCodeGenerator):

  def __init__(self,
               onnx_ver=onnx.defs.onnx_opset_version(),
               torch_ver=torch.__version__):
    super(ResizeOpCodeGenerator, self).__init__(onnx_ver, torch_ver)

  def gen(self, node, value_infos, initializers, rename_helper, tensor_inplace):
    attr_value_dict = self.get_attr_value_dict(node)
    inputs_str, outputs_str = self.gen_input_output_string(
        node, initializers, rename_helper, tensor_inplace)
    scales, sizes = None, None
    if len(node.input) == 4:
      sizes = tuple(onnx.numpy_helper.to_array(initializers[node.input[3]])[2:])
    else:
      scales = tuple(
          onnx.numpy_helper.to_array(initializers[node.input[2]])[2:])

    align_corners = None
    if attr_value_dict["coordinate_transformation_mode"].decode(
    ) == "align_corners":
      align_corners = True
    mode = attr_value_dict['mode'].decode()
    d = len(value_infos[node.input[0]].type.tensor_type.shape.dim) - 2
    if mode == "linear":
      if d == 1:
        pass
      elif d == 2:
        mode = "bilinear"
      elif d == 3:
        mode = "bicubic"
      elif d == 4:
        mode = "trilinear"
      else:
        raise NotImplementedError
    params_str = self.gen_params_str(
        size=sizes,
        scale_factor=scales,
        mode=f"'{mode}'",
        align_corners=align_corners,
        recompute_scale_factor=scales is not None,
    )
    init_str, forward_str = [], []

    forward_str.append(
        f"{outputs_str[0]} = F.interpolate({inputs_str[0]}, **{{{params_str}}})"
    )
    return {"init": init_str, "forward": forward_str}



import onnx
import torch

from onnx_pytorch.op_code_generators import OpCodeGenerator


class BitShiftOpCodeGenerator(OpCodeGenerator):

  def __init__(self,
               onnx_ver=onnx.defs.onnx_opset_version(),
               torch_ver=torch.__version__):
    super(BitShiftOpCodeGenerator, self).__init__(onnx_ver, torch_ver)

  def gen(self, node, value_infos, initializers, rename_helper, tensor_inplace):
    attr_value_dict = self.get_attr_value_dict(node)
    inputs_str, outputs_str = self.gen_input_output_string(
        node, initializers, rename_helper, tensor_inplace)
    init_str, forward_str = [], []
    direction = "<<" if attr_value_dict["direction"] == b"LEFT" else ">>"
    forward_str.append(
        f"{outputs_str[0]} = {f' {direction} '.join(inputs_str)}")
    return {"init": init_str, "forward": forward_str}



import onnx
import torch
from onnx.numpy_helper import to_array

from onnx_pytorch.op_code_generators import OpCodeGenerator


class ConstantOfShapeOpCodeGenerator(OpCodeGenerator):

  def __init__(self,
               onnx_ver=onnx.defs.onnx_opset_version(),
               torch_ver=torch.__version__):
    super(ConstantOfShapeOpCodeGenerator, self).__init__(onnx_ver, torch_ver)

  def gen(self, node, value_infos, initializers, rename_helper, tensor_inplace):
    attr_value_dict = self.get_attr_value_dict(node)
    inputs_str, outputs_str = self.gen_input_output_string(
        node, initializers, rename_helper, tensor_inplace)
    init_str, forward_str = [], []
    dtype = "float32"
    val = 0.
    if "value" in attr_value_dict:
      array = to_array(attr_value_dict["value"])
      dtype = array.dtype
      val = array[0]
    forward_str.append(
        f"{outputs_str[0]} = torch.Tensor().new_full(size={inputs_str[0]}.tolist(), fill_value={val}, dtype=torch.{dtype})"
    )
    return {"init": init_str, "forward": forward_str}



import onnx
import torch

from onnx_pytorch.op_code_generators import OpCodeGenerator


class PadOpCodeGenerator(OpCodeGenerator):

  def __init__(self,
               onnx_ver=onnx.defs.onnx_opset_version(),
               torch_ver=torch.__version__):
    super(PadOpCodeGenerator, self).__init__(onnx_ver, torch_ver)

  def gen(self, node, value_infos, initializers, rename_helper, tensor_inplace):
    attr_value_dict = self.get_attr_value_dict(node)
    inputs_str, outputs_str = self.gen_input_output_string(
        node, initializers, rename_helper, tensor_inplace)
    init_str, forward_str = [], []
    mode = attr_value_dict.get("mode", b"constant")
    value = 0.
    if mode == b"constant":
      if len(node.input) == 3:
        value = onnx.numpy_helper.to_array(initializers[node.input[2]])[0]
    d = len(value_infos[node.input[0]].type.tensor_type.shape.dim) - 2
    pads = initializers.get(node.input[1], None)
    assert pads is not None, "Currently PadOpCodeGenerator only support all of [pads] is in initializers."
    pads = onnx.numpy_helper.to_array(pads)
    pt_pads = [0, 0] * d
    for i in range(d):
      pt_pads[2 * (d - i - 1)] = pads[2 + i]
      pt_pads[2 * (d - i - 1) + 1] = pads[d + 2 + 2 + i]
    forward_str.append(
        f"{outputs_str[0]} = F.pad({inputs_str[0]}, {pt_pads.__repr__()}, \"{mode.decode()}\", {value})"
    )
    return {"init": init_str, "forward": forward_str}



import argparse
import logging
import os
import re
import shutil
from collections import Counter

import numpy as np
import onnx
from onnx.mapping import TENSOR_TYPE_TO_NP_TYPE
from onnx.numpy_helper import to_array
from onnxruntime.tools.symbolic_shape_infer import SymbolicShapeInference

from onnx_pytorch.code_gen_template import CodeGenTemplate
from onnx_pytorch.op_code_generators import *


class RenameHelper:

  def __init__(self, simplify_names=False):
    self.simplify_names = simplify_names

    self.tensor_name_mapping = {}
    self.tensor_name_counter = Counter()
    self.node_name_mapping = {}
    self.node_name_counter = Counter()

    self.tensor_counter = 0
    self.node_counter = Counter()

    self.init_name_set = set()
    self.sim_tensor_name_set = set()

  def get_tensor_name(self, tensor_name):
    if self.simplify_names:
      return self.get_simplify_tensor_name(tensor_name)
    return tensor_name

  def get_node_name(self, node_name, op_type):
    if self.simplify_names:
      return self.get_simplify_node_name(node_name, op_type)
    return f"n_{node_name}"

  def get_simplify_node_name(self, node_name, op_type):
    idx = self.node_counter[op_type]
    self.node_counter[op_type] += 1
    self.node_name_mapping[node_name] = f"n_{op_type}_{idx}"
    return self.node_name_mapping[node_name]

  def get_simplify_tensor_name(self, tensor_name):
    if tensor_name in self.tensor_name_mapping:
      return self.tensor_name_mapping[tensor_name]
    suffix = self.tensor_counter
    self.tensor_counter += 1
    sim_tensor_name = f"t_{suffix}"
    # Only entry when use_random
    while sim_tensor_name in self.sim_tensor_name_set:
      sim_tensor_name = f"t_{suffix}"
    self.sim_tensor_name_set.add(sim_tensor_name)
    self.tensor_name_mapping[tensor_name] = sim_tensor_name
    return self.tensor_name_mapping[tensor_name]


class ModelCodeGenerator:

  def __init__(self,
               onnx_model=None,
               output_dir=None,
               simplify_names=False,
               tensor_inplace=False,
               continue_on_error=False):
    self.onnx_model = onnx_model
    self.output_dir = output_dir
    self.rename_helper = RenameHelper(simplify_names)
    self.tensor_inplace = tensor_inplace
    self.continue_on_error = continue_on_error
    self.init_parts = []
    self.forward_parts = []

  def add_init_part(self, m):
    if type(m) in (list, tuple, set):
      self.init_parts.extend(m)
    else:
      self.init_parts.append(m)

  def add_forward_part(self, m):
    if type(m) in (list, tuple, set):
      self.forward_parts.extend(m)
    else:
      self.forward_parts.append(m)

  def add_forward_return(self, outputs_value_infos):
    return_list = [
        f"{self.rename_helper.get_tensor_name(o.name)}"
        for o in outputs_value_infos
    ]
    self.forward_parts.append(f"return {', '.join(return_list)}")

  def add_forward_input(self, inputs_value_infos):
    initializer_names = {i.name for i in self.onnx_model.graph.initializer}
    return_list = [
        f"{self.rename_helper.get_tensor_name(i.name)}"
        for i in inputs_value_infos
        if i.name not in initializer_names
    ]
    if len(return_list) == 1:
      self.forward_parts.append(f"{return_list[0]}, = inputs")
    else:
      self.forward_parts.append(f"{', '.join(return_list)} = inputs")

  def gen_model_code(self):
    return CodeGenTemplate.model(model_init='''
    '''.join(self.init_parts),
                                 model_forward='''
    '''.join(self.forward_parts),
                                 test_run_model=self.gen_test_run_model_code())

  def gen_test_run_model_code(self):
    numpy_input_str = []
    initializer_names = {i.name for i in self.onnx_model.graph.initializer}
    for i in self.onnx_model.graph.input:
      if i.name in initializer_names:
        continue
      dtype = TENSOR_TYPE_TO_NP_TYPE[i.type.tensor_type.elem_type]
      shape = []
      for d in i.type.tensor_type.shape.dim:
        if d.dim_param != "":
          shape.append(1)
        else:
          shape.append(d.dim_value)
      if shape:
        numpy_input_str.append(
            f"torch.from_numpy(np.random.randn(*{[s if s > 1 else 1 for s in shape].__repr__()}).astype(np.{dtype.name}))"
        )
      else:
        numpy_input_str.append(
            f"torch.from_numpy(np.random.randn(1).astype(np.{dtype.name}))")
    test_run_model = [
        f'''@torch.no_grad()
def test_run_model(inputs=[{', '.join(numpy_input_str)}]):''',
        "model = Model()", "model.eval()", "print(model)"
    ]
    test_run_model.extend(["rs = model(*inputs)", "print(rs)", "return rs"])
    return '''
  '''.join(test_run_model)

  def preprocess_onnx_model(self):
    self.onnx_model.graph.ClearField("value_info")
    for n in self.onnx_model.graph.node:
      inputs, outputs = [], []
      for ls, f in ((inputs, n.input), (outputs, n.output)):
        for i in f:
          new_i = re.sub("[:/.]", "_", i)
          ls.append(f"t_{new_i}")
          if i != ls[-1] and not self.rename_helper.simplify_names:
            logging.warning(f"Tensor name {i} is changed to {ls[-1]}.")
          self.rename_helper.tensor_name_counter[ls[-1]] += 1

      n.ClearField("input")
      n.input.extend(inputs)
      n.ClearField("output")
      n.output.extend(outputs)

      old_name = n.name
      n.name = re.sub("[:/.]", "_", n.name)
      n.name = f"n_{n.name}"
      if old_name != n.name and not self.rename_helper.simplify_names:
        logging.warning(f"Node name {old_name} is changed to {n.name}.")
      self.rename_helper.node_name_counter[n.name] += 1

    for f in (self.onnx_model.graph.input, self.onnx_model.graph.output,
              self.onnx_model.graph.initializer):
      for i in f:
        old_name = i.name
        i.name = re.sub("[:/.]", "_", i.name)
        i.name = f"t_{i.name}"
        if old_name != i.name and not self.rename_helper.simplify_names:
          logging.warning(f"Tensor name {i.name} is changed to {i.name}.")
        self.rename_helper.tensor_name_counter[i.name] += 1

    # TODO how to deal with custom op?
    model = self.onnx_model
    try:
      model = SymbolicShapeInference.infer_shapes(model, 2**31 - 1, True, True,
                                                  0)
    except:
      logging.warning("Shape infer by onnxruntime failed.")
    onnx.save(model, os.path.join(self.output_dir, "tmp_processed.onnx"))
    self.onnx_model = model

  def run(self):
    self.preprocess_onnx_model()
    initializers = {i.name: i for i in self.onnx_model.graph.initializer}
    input_value_infos = {i.name: i for i in self.onnx_model.graph.input}
    output_value_infos = {i.name: i for i in self.onnx_model.graph.output}
    value_infos = {}
    value_infos.update(input_value_infos)
    value_infos.update(output_value_infos)
    value_infos.update({i.name: i for i in self.onnx_model.graph.value_info})

    for i in self.onnx_model.graph.initializer:
      self.rename_helper.get_tensor_name(i.name)

    self.add_forward_input(self.onnx_model.graph.input)
    for n in self.onnx_model.graph.node:
      op_code_gen = get_op_code_generator(n.op_type)
      if op_code_gen is None:
        if self.continue_on_error:
          self.add_forward_part(n.__repr__())
          logging.warning(f"OpCodeGenerator is unimplemented for {n.op_type}. "
                          "Please modify this part by manual later.")
        else:
          raise NotImplementedError(
              f"OpCodeGenerator is unimplemented for {n.op_type}.")
      gened = op_code_gen.gen(n, value_infos, initializers, self.rename_helper,
                              self.tensor_inplace)
      self.add_init_part(gened["init"])
      self.add_forward_part(gened["forward"])
    self.add_forward_return(self.onnx_model.graph.output)

    gened_code = self.gen_model_code()
    print(gened_code)
    with open(os.path.join(self.output_dir, "model.py"), "w") as f:
      f.write(gened_code)
    shutil.rmtree(os.path.join(self.output_dir, "variables"),
                  ignore_errors=True)
    os.makedirs(os.path.join(self.output_dir, "variables"))
    for k, v in initializers.items():
      np.save(
          os.path.join(self.output_dir, "variables",
                       f"{self.rename_helper.get_tensor_name(k)}.npy"),
          to_array(v))


def gen(onnx_model,
        output_dir,
        overwrite=False,
        tensor_inplace=False,
        simplify_names=False,
        continue_on_error=False):
  kwargs = {
      "output_dir": output_dir,
      "simplify_names": simplify_names,
      "tensor_inplace": tensor_inplace,
      "continue_on_error": continue_on_error,
  }
  if type(onnx_model) == onnx.ModelProto:
    kwargs["onnx_model"] = onnx_model
  else:
    assert os.path.exists(
        onnx_model), f"ONNX model {onnx_model} does not exist."
    assert os.path.isfile(onnx_model), f"{onnx_model} is not a file."
    assert os.path.exists(
        output_dir
    ) and overwrite is not True, f"{output_dir} is not empty and overwrite is not True."
    assert os.path.isdir(output_dir), f"{output_dir} is not directory."
    kwargs["onnx_model"] = onnx.load(onnx_model)
  if overwrite:
    shutil.rmtree(output_dir, ignore_errors=True)
    os.makedirs(output_dir)
  ModelCodeGenerator(**kwargs).run()


def main():
  debug = True
  parser = argparse.ArgumentParser()
  parser.add_argument("--onnx_model_path",
                      default=None,
                      type=str,
                      required=not debug,
                      help="The onnx model path.")
  parser.add_argument("--output_dir",
                      default=None,
                      type=str,
                      required=not debug,
                      help="The output dir")
  parser.add_argument("--overwrite",
                      default=False,
                      type=bool,
                      help="Should overwrite the output dir.")
  parser.add_argument("--tensor_inplace",
                      default=False,
                      type=bool,
                      help="Try best to inplace tensor.")
  parser.add_argument("--continue_on_error",
                      default=False,
                      type=bool,
                      help="Continue on error.")
  parser.add_argument(
      "--simplify_names",
      default=False,
      type=int,
      help="Use indexing shorten name instead of original name.")
  args = parser.parse_args()

  gen(onnx_model=args.onnx_model_path,
      output_dir=args.output_dir,
      overwrite=args.overwrite,
      tensor_inplace=args.tensor_inplace,
      simplify_names=args.simplify_names,
      continue_on_error=args.continue_on_error)


if __name__ == '__main__':
  main()



import onnx
import torch

from onnx_pytorch.op_code_generators import OpCodeGenerator


class ClipOpCodeGenerator(OpCodeGenerator):

  def __init__(self,
               onnx_ver=onnx.defs.onnx_opset_version(),
               torch_ver=torch.__version__):
    super(ClipOpCodeGenerator, self).__init__(onnx_ver, torch_ver)

  def gen(self, node, value_infos, initializers, rename_helper, tensor_inplace):
    attr_value_dict = self.get_attr_value_dict(node)
    inputs_str, outputs_str = self.gen_input_output_string(
        node, initializers, rename_helper, tensor_inplace)
    min = attr_value_dict.get("min", "float(\"-inf\")")
    max = attr_value_dict.get("max", "float(\"inf\")")
    if len(inputs_str) == 1:
      inputs_str.append(min)
    if len(inputs_str) < 3:
      inputs_str.append(max)
    init_str, forward_str = [], []
    forward_str.append(
        f"{outputs_str[0]} = torch.clip({', '.join(inputs_str)})")
    return {"init": init_str, "forward": forward_str}



import onnx
import onnx.numpy_helper
import torch

from onnx_pytorch.op_code_generators import OpCodeGenerator


class ConvTransposeOpCodeGenerator(OpCodeGenerator):

  def __init__(self,
               onnx_ver=onnx.defs.onnx_opset_version(),
               torch_ver=torch.__version__):
    super(ConvTransposeOpCodeGenerator, self).__init__(onnx_ver, torch_ver)

  def gen(self, node, value_infos, initializers, rename_helper, tensor_inplace):
    attr_value_dict = self.get_attr_value_dict(node)
    inputs_str, outputs_str = self.gen_input_output_string(
        node, initializers, rename_helper, tensor_inplace)

    d = len(value_infos[node.input[0]].type.tensor_type.shape.dim) - 2
    input_size = [
        d.dim_value
        for d in value_infos[node.input[0]].type.tensor_type.shape.dim
    ][2:]
    assert (d in (1, 2, 3))

    weights = onnx.numpy_helper.to_array(initializers[node.input[1]])
    padding = [0] * d
    output_padding = [0] * d
    stride = attr_value_dict.get("strides", [1] * d)
    kernel_shape = weights.shape[2:]
    dilation = attr_value_dict.get("dilations", [1] * d)
    if "pads" in attr_value_dict:
      padding = [attr_value_dict["pads"][i] for i in range(d)]
    if "output_padding" in attr_value_dict:
      output_padding = [attr_value_dict["output_padding"][i] for i in range(d)]
    if "output_shape" in attr_value_dict:
      output_shape = attr_value_dict["output_shape"]
      total_padding = [0] * d

      # total_padding[i] = stride[i] * (input_size[i] - 1) + output_padding[i] + ((kernel_shape[i] - 1) * dilations[i] + 1) - output_shape[i]
      # If (auto_pads == SAME_UPPER): pads[start_i] = total_padding[i]/2; pads[end_i] = total_padding[i] - (total_padding[i]/2)
      # Else: pads[start_i] = total_padding[i] - (total_padding[i]/2); pads[end_i] = (total_padding[i]/2).

      for i in range(d):
        total_padding[i] = stride[i] * (
            input_size[i] - 1) + output_padding[i] + (
                (kernel_shape[i] - 1) * dilation[i] + 1) - output_shape[i]
        assert total_padding[
            i] % 2 == 0, "Padding for ConvTranspose should be even."
        padding[i] = total_padding[i] // 2
    params_str = self.gen_params_str(groups=attr_value_dict["group"],
                                     dilation=dilation,
                                     out_channels=weights.shape[1],
                                     padding=padding,
                                     output_padding=output_padding,
                                     kernel_size=weights.shape[2:],
                                     stride=stride,
                                     in_channels=weights.shape[0],
                                     bias=len(node.input) > 2)

    nn_name = f"ConvTranspose{d}d"
    node_name = rename_helper.get_node_name(node.name, node.op_type)
    init_str, forward_str = [], []
    init_str.append(f"self.{node_name} = nn.{nn_name}(**{{{params_str}}})")
    init_str.append(f"self.{node_name}.weight.data = {inputs_str[1]}")
    if len(node.input) > 2:
      init_str.append(f"self.{node_name}.bias.data = {inputs_str[2]}")

    forward_str.append(f"{outputs_str[0]} = self.{node_name}({inputs_str[0]})")

    return {"init": init_str, "forward": forward_str}
